@@ -8,19 +8,25 @@ extern crate pathdiff;
 extern crate serde_yaml;
 
 use indexmap::{IndexMap, IndexSet};
+use itertools::Itertools;
 use std::collections::{HashMap, HashSet};
 use std::fs::read_to_string;
 use std::path::{Path, PathBuf};
 use std::time::Instant;
 
-use anyhow::{Context as _, Result};
+use anyhow::{Context as _, Error, Result};
 use serde::{Deserialize, Deserializer};
 
 use treestate::{FileState, TreeState};
 
 use super::download::Download;
+use super::model::CustomBuild;
 use super::nested_env::{Env, EnvKey, MergeOption};
 use super::{Context, ContextBag, Dependency, Module, Rule, Task};
+use crate::serde_bool_helpers::default_as_false;
+
+mod import;
+use import::Import;
 
 pub type FileTreeState = TreeState<FileState, PathBuf>;
 
@@ -41,17 +47,17 @@ struct YamlFile {
     module: Option<Option<Vec<YamlModule>>>,
     #[serde(default, deserialize_with = "deserialize_some")]
     app: Option<Option<Vec<YamlModule>>>,
-    import: Option<Vec<String>>,
+    import: Option<Vec<Import>>,
     subdirs: Option<Vec<String>>,
     defaults: Option<HashMap<String, YamlModule>>,
     #[serde(skip)]
     filename: Option<PathBuf>,
     #[serde(skip)]
-    included_from: Option<PathBuf>,
-    #[serde(skip)]
     doc_idx: Option<usize>,
     #[serde(skip)]
     included_by: Option<usize>,
+    #[serde(skip)]
+    import_root: Option<ImportRoot>,
 }
 
 #[derive(Debug, Serialize, Deserialize)]
@@ -85,16 +91,22 @@ enum StringOrVecString {
 #[derive(Debug, Serialize, Deserialize)]
 struct YamlModule {
     name: Option<String>,
-    context: Option<String>,
+    context: Option<StringOrVecString>,
     depends: Option<Vec<StringOrMapString>>,
     selects: Option<Vec<String>>,
     uses: Option<Vec<String>>,
     disable: Option<Vec<String>>,
+    #[serde(default = "default_as_false")]
+    notify_all: bool,
     sources: Option<Vec<StringOrMapString>>,
+    build: Option<CustomBuild>,
     env: Option<YamlModuleEnv>,
     blocklist: Option<Vec<String>>,
     allowlist: Option<Vec<String>>,
     download: Option<Download>,
+    srcdir: Option<PathBuf>,
+    #[serde(default = "default_as_false")]
+    is_build_dep: bool,
     #[serde(skip)]
     is_binary: bool,
 }
@@ -108,14 +120,29 @@ impl YamlModule {
             selects: None,
             uses: None,
             disable: None,
+            notify_all: false,
             sources: None,
+            srcdir: None,
+            build: None,
             env: None,
             blocklist: None,
             allowlist: None,
             download: None,
+            is_build_dep: false,
             is_binary,
         }
     }
+
+    fn get_contexts(&self) -> Vec<Option<&String>> {
+        if let Some(contexts) = &self.context {
+            match contexts {
+                StringOrVecString::Single(single) => vec![Some(&single)],
+                StringOrVecString::List(list) => list.iter().map(|x| Some(x)).collect_vec(),
+            }
+        } else {
+            return vec![None];
+        }
+    }
 }
 
 #[derive(Debug, Serialize, Deserialize)]
@@ -136,7 +163,7 @@ struct YamlModuleEnv {
 //     Ok(data)
 // }
 
-fn process_removes(strings: &mut Vec<Dependency>) {
+fn process_removes(strings: &mut Vec<Dependency<String>>) {
     let removals = strings
         .iter()
         .filter(|x| x.get_name().starts_with("-"))
@@ -146,43 +173,79 @@ fn process_removes(strings: &mut Vec<Dependency>) {
     strings.retain(|x| !(x.get_name().starts_with("-") || removals.contains(&x.get_name()[..])));
 }
 
-pub fn dependency_from_string(dep_name: &String) -> Dependency {
+pub fn dependency_from_string(dep_name: &String) -> Dependency<String> {
     match dep_name.as_bytes()[0] {
         b'?' => Dependency::Soft(dep_name[1..].to_string()),
         _ => Dependency::Hard(dep_name.clone()),
     }
 }
 
-pub fn dependency_from_string_if(dep_name: &String, other: &String) -> Dependency {
+pub fn dependency_from_string_if(dep_name: &String, other: &String) -> Dependency<String> {
     match dep_name.as_bytes()[0] {
         b'?' => Dependency::IfThenSoft(other.clone(), dep_name[1..].to_string()),
         _ => Dependency::IfThenHard(other.clone(), dep_name.clone()),
     }
 }
 
-fn load_all<'a>(
-    filename: &PathBuf,
-    index_start: usize,
-    included_by: Option<usize>,
-) -> Result<Vec<YamlFile>> {
-    let file = read_to_string(filename)
-        .with_context(|| format!("while reading {}", filename.display()))?;
-
-    let docs: Vec<&str> = file.split("\n---\n").collect();
+fn load_all<'a>(file_include: &FileInclude, index_start: usize) -> Result<Vec<YamlFile>> {
+    let filename = &file_include.filename;
+    let file = read_to_string(filename).with_context(|| format!("{:?}", filename))?;
 
     let mut result = Vec::new();
-    for (n, doc) in docs.iter().enumerate() {
-        let mut parsed: YamlFile = serde_yaml::from_str(doc)
-            .with_context(|| format!("while parsing {}", filename.display()))?;
+    for (n, doc) in serde_yaml::Deserializer::from_str(&file).enumerate() {
+        let mut parsed =
+            YamlFile::deserialize(doc).with_context(|| format!("{}", filename.display()))?;
         parsed.filename = Some(filename.clone());
         parsed.doc_idx = Some(index_start + n);
-        parsed.included_by = included_by;
+        parsed.included_by = file_include.included_by_doc_idx;
+        parsed.import_root = file_include.import_root.clone();
         result.push(parsed);
     }
 
     Ok(result)
 }
 
+#[derive(Hash, Debug, PartialEq, Eq, Clone)]
+struct ImportRoot(PathBuf);
+impl ImportRoot {
+    fn path(&self) -> &Path {
+        self.0.as_path()
+    }
+}
+
+#[derive(Hash, Debug, PartialEq, Eq)]
+struct FileInclude {
+    filename: PathBuf,
+    included_by_doc_idx: Option<usize>,
+    import_root: Option<ImportRoot>,
+}
+
+impl FileInclude {
+    fn new(
+        filename: PathBuf,
+        included_by_doc_idx: Option<usize>,
+        import_root: Option<ImportRoot>,
+    ) -> Self {
+        FileInclude {
+            filename,
+            included_by_doc_idx,
+            import_root,
+        }
+    }
+
+    fn new_import(filename: PathBuf, included_by_doc_idx: Option<usize>) -> Self {
+        // TODO: (opt) Cow import_root?
+        let import_root = Some(ImportRoot(PathBuf::from(
+            filename.parent().as_ref().unwrap(),
+        )));
+        FileInclude {
+            filename,
+            included_by_doc_idx,
+            import_root,
+        }
+    }
+}
+
 pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeState)> {
     let mut contexts = ContextBag::new();
     let start = Instant::now();
@@ -194,23 +257,17 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
     // when reading files, any "subdir" will be converted to "subdir/laze.yml", then added to the
     // set.
     // using an IndexSet so files can only be added once
-    let mut filenames: IndexSet<(_, Option<usize>)> = IndexSet::new();
-    filenames.insert((PathBuf::from(filename), None));
+    let mut filenames: IndexSet<FileInclude> = IndexSet::new();
+    filenames.insert(FileInclude::new(PathBuf::from(filename), None, None));
 
     let mut filenames_pos = 0;
     while filenames_pos < filenames.len() {
-        let (filename, included_by_doc_idx) = filenames.get_index(filenames_pos).unwrap();
-        let filename = filename.clone();
+        let include = filenames.get_index(filenames_pos).unwrap();
+        let filename = include.filename.clone();
         let new_index_start = yaml_datas.len();
 
-        let included_by = if let Some(doc_idx) = included_by_doc_idx {
-            Some(*doc_idx)
-        } else {
-            None
-        };
-
         // load all yaml documents from filename, append to yaml_datas
-        yaml_datas.extend(load_all(&filename, new_index_start, included_by)?.drain(..));
+        yaml_datas.extend(load_all(&include, new_index_start)?.drain(..));
         filenames_pos += 1;
 
         let new_index_end = yaml_datas.len();
@@ -224,7 +281,22 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
                 // collect subdirs, add do filenames list
                 for subdir in subdirs {
                     let sub_file = Path::new(&relpath).join(subdir).join("laze.yml");
-                    filenames.insert((sub_file, new.doc_idx));
+                    filenames.insert(FileInclude::new(
+                        sub_file,
+                        new.doc_idx,
+                        new.import_root.clone(),
+                    ));
+                }
+            }
+            if let Some(imports) = &new.import {
+                for import in imports {
+                    // TODO: `import.handle()` does the actual git checkout (or whatever
+                    // import action), so probably better handling of any errors is
+                    // in order.
+                    filenames.insert(FileInclude::new_import(
+                        import.handle(build_dir)?,
+                        new.doc_idx,
+                    ));
                 }
             }
         }
@@ -235,7 +307,8 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
         contexts: &mut ContextBag,
         is_builder: bool,
         filename: &PathBuf,
-    ) {
+        import_root: &Option<ImportRoot>,
+    ) -> Result<(), Error> {
         let context_name = &context.name;
         let context_parent = match &context.parent {
             Some(x) => x.clone(),
@@ -261,7 +334,7 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
                 ),
                 is_builder,
             )
-            .unwrap();
+            .with_context(|| format!("{:?}: adding context \"{}\"", &filename, &context_name))?;
         context_.env = context.env.clone();
         if let Some(rules) = &context.rule {
             context_.rules = Some(IndexMap::new());
@@ -276,9 +349,6 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
             }
         }
         context_.var_options = context.var_options.clone();
-        if let Some(tasks) = &context.tasks {
-            context_.tasks = Some(tasks.clone());
-        }
         // populate "early env"
         let relpath = {
             let relpath = filename.parent().unwrap().to_str().unwrap();
@@ -292,9 +362,27 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
         context_
             .env_early
             .insert("relpath".into(), EnvKey::Single(relpath));
-        context_
-            .env_early
-            .insert("root".into(), EnvKey::Single(".".into()));
+        if let Some(import_root) = import_root {
+            context_.env_early.insert(
+                "root".into(),
+                EnvKey::Single(import_root.path().to_str().unwrap().into()),
+            );
+        } else {
+            context_
+                .env_early
+                .insert("root".into(), EnvKey::Single(".".into()));
+        }
+
+        if let Some(tasks) = &context.tasks {
+            let flattened_early_env = crate::nested_env::flatten(&context_.env_early);
+            context_.tasks = Some(
+                tasks
+                    .iter()
+                    .map(|(name, task)| (name.clone(), task.with_env(&flattened_early_env)))
+                    .collect(),
+            )
+        }
+
         context_.apply_early_env();
 
         context_.defined_in = Some(filename.clone());
@@ -310,6 +398,8 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
                     .collect::<Vec<_>>(),
             );
         }
+
+        Ok(())
     }
 
     fn init_module(
@@ -317,12 +407,26 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
         context: Option<&String>,
         is_binary: bool,
         filename: &PathBuf,
+        import_root: &Option<ImportRoot>,
         defaults: Option<&Module>,
     ) -> Module {
         let relpath = filename.parent().unwrap().to_str().unwrap().to_string();
         let name = match name {
             Some(name) => name.clone(),
-            None => relpath.clone(),
+            None => {
+                if let Some(import_root) = import_root {
+                    filename
+                        .parent()
+                        .unwrap()
+                        .strip_prefix(import_root.path())
+                        .unwrap()
+                        .to_str()
+                        .unwrap()
+                        .to_string()
+                } else {
+                    relpath.clone()
+                }
+            }
         };
 
         let mut module = match defaults {
@@ -339,17 +443,21 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
 
     fn convert_module(
         module: &YamlModule,
+        context: Option<&String>,
         is_binary: bool,
         filename: &PathBuf,
+        import_root: &Option<ImportRoot>,
         defaults: Option<&Module>,
         build_dir: &Path,
-    ) -> Module {
+    ) -> Result<Module, Error> {
         let relpath = filename.parent().unwrap().to_str().unwrap().to_string();
+
         let mut m = init_module(
             &module.name,
-            module.context.as_ref(),
+            context,
             is_binary,
             filename,
+            import_root,
             defaults,
         );
 
@@ -409,10 +517,18 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
                     m.disable.as_mut().unwrap().push(dep_name.clone());
                 }
             } else {
-                eprintln!("laze: warning: \"disable\" ignored on regular modules!");
+                return Err(anyhow!(
+                    "{:?}: module {}: \"disable\" not allowed on regular (non-app) modules!",
+                    m.defined_in.unwrap(),
+                    m.name
+                ));
             }
         }
 
+        if module.notify_all {
+            m.notify_all = true;
+        }
+
         // if a module name starts with "-", remove it from the list, also the
         // same name without "-".
         // this allows adding e.g., a dependency in "default: ...", but removing
@@ -483,42 +599,56 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
         }
 
         m.download = module.download.clone();
-        let (srcdir, build_deps) = if let Some(download) = &m.download {
+        let srcdir = if let Some(download) = &m.download {
             let srcdir = download.srcdir(build_dir, &m);
             let tagfile = download.tagfile(&srcdir);
 
-            // Create "build_deps" list in export env of this module.
-            // This will later be picked up and added as dependency for
-            // every dependent module's source file.
-            m.env_export.insert(
-                "build_deps".into(),
-                // TODO: why is this so ugly and still unsafe (e.g., in Windows)
-                EnvKey::List(im::vector![tagfile
-                    .clone()
-                    .into_os_string()
-                    .into_string()
-                    .unwrap()]),
-            );
-            (srcdir, Some(vec![tagfile]))
+            m.add_build_dep_file(&tagfile);
+
+            // if a module has downloaded files, always consider it to be a
+            // build dependency, as all dependees / users might include e.g.,
+            // downloaded headers.
+            m.is_build_dep = true;
+
+            srcdir
         } else {
-            (PathBuf::from(relpath.clone()), None)
+            PathBuf::from(relpath.clone())
         };
 
-        m.srcdir = Some(srcdir);
-        m.build_deps = build_deps;
+        m.build = module.build.clone();
+
+        if m.download.is_none() {
+            // if a module has downloaded source, it is already a build dependency
+            // for dependees / users. Otherwise, do what the user thinks.
+            m.is_build_dep = module.is_build_dep;
+        }
+
+        m.srcdir = module
+            .srcdir
+            .as_ref()
+            .map_or(Some(srcdir), |s| Some(PathBuf::from(s)));
 
         // populate "early env"
         m.env_early
             .insert("relpath".into(), EnvKey::Single(relpath));
-        m.env_early
-            .insert("root".into(), EnvKey::Single(".".into()));
+        if let Some(import_root) = import_root {
+            m.env_early.insert(
+                "root".into(),
+                EnvKey::Single(import_root.path().to_str().unwrap().into()),
+            );
+        } else {
+            m.env_early
+                .insert("root".into(), EnvKey::Single(".".into()));
+        }
         m.env_early.insert(
             "srcdir".into(),
             EnvKey::Single(m.srcdir.as_ref().unwrap().to_str().unwrap().into()),
         );
+
+        m.env_local = crate::nested_env::merge(m.env_local, m.env_early.clone());
         m.apply_early_env();
 
-        m
+        Ok(m)
     }
 
     // collect and convert contexts
@@ -533,7 +663,8 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
                         &mut contexts,
                         *is_builder,
                         &data.filename.as_ref().unwrap(),
-                    );
+                        &data.import_root,
+                    )?;
                 }
             }
         }
@@ -541,7 +672,7 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
 
     // after this, there's a default context, context relationships and envs have been set up.
     // modules can now be processed.
-    contexts.finalize();
+    contexts.finalize()?;
 
     // for context in &contexts.contexts {
     //     if let Some(env) = &context.env {
@@ -571,13 +702,29 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
         // determine "defaults: module: ..." from yaml document
         let mut module_defaults = if let Some(defaults) = &data.defaults {
             if let Some(module_defaults) = defaults.get(key) {
-                Some(convert_module(
-                    &module_defaults,
-                    is_binary,
-                    &data.filename.as_ref().unwrap(),
-                    subdir_defaults,
-                    build_dir,
-                ))
+                let context =
+                    &module_defaults
+                        .context
+                        .as_ref()
+                        .map_or(None, |context| match context {
+                            StringOrVecString::List(_) => {
+                                panic!("module defaults with context _list_")
+                            }
+                            StringOrVecString::Single(context) => Some(context),
+                        });
+
+                Some(
+                    convert_module(
+                        &module_defaults,
+                        *context,
+                        is_binary,
+                        &data.filename.as_ref().unwrap(),
+                        &data.import_root,
+                        subdir_defaults,
+                        build_dir,
+                    )
+                    .unwrap(),
+                )
             } else {
                 None
             }
@@ -615,34 +762,38 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
             if let Some(module_list) = list {
                 if let Some(module_list) = module_list {
                     for module in module_list {
-                        contexts
-                            .add_module(convert_module(
+                        for context in module.get_contexts() {
+                            contexts.add_module(convert_module(
                                 &module,
+                                context,
                                 *is_binary,
                                 &data.filename.as_ref().unwrap(),
+                                &data.import_root,
                                 if *is_binary {
                                     app_defaults.as_ref()
                                 } else {
                                     module_defaults.as_ref()
                                 },
                                 build_dir,
-                            ))
-                            .unwrap();
+                            )?)?;
+                        }
                     }
                 } else {
                     if *is_binary {
                         // if an app list is empty, add a default entry.
                         // this allows a convenient file only containing "app:"
                         let module = YamlModule::default(*is_binary);
-                        contexts
-                            .add_module(convert_module(
+                        for context in module.get_contexts() {
+                            contexts.add_module(convert_module(
                                 &module,
+                                context,
                                 *is_binary,
                                 &data.filename.as_ref().unwrap(),
+                                &data.import_root,
                                 app_defaults.as_ref(),
                                 build_dir,
-                            ))
-                            .unwrap();
+                            )?)?;
+                        }
                     }
                 }
             }
@@ -656,7 +807,7 @@ pub fn load(filename: &Path, build_dir: &Path) -> Result<(ContextBag, FileTreeSt
     );
 
     let start = Instant::now();
-    let treestate = FileTreeState::new(filenames.iter().map(|(path, _)| path));
+    let treestate = FileTreeState::new(filenames.iter().map(|include| &include.filename));
     println!(
         "laze: stat'ing {} files took {:?}",
         filenames.len(),
