@@ -7,7 +7,7 @@ cleanup
 ${LAZE} task echo foo | grep -s "^foo$"
 ${LAZE} task foobar | grep -s "^foobar$"
 ${LAZE} task --global --app subdir --builder default vars \
-         |grep -s "^relpath=subdir relroot=.. out=build/default/subdir/subdir.elf builder=default$"
+         |grep -s "^relpath=\. relroot=\.\. out=build/default/subdir/subdir.elf builder=default$"
 
 echo TEST_OK
 
